#
# Perl extension to read and process NEdit configuration and import files
#
# Copyright (c) 2000 Joor Loohuis. All rights reserved.
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# revision history:
# date		author/comment
# Sep 26, 2001	v0.23, Joor Loohuis
# 		fixed a write bug in the languageMode formatting code
# 2000/10/23	v0.22, Joor Loohuis
#		fixed a critical bug in the pattern formatting code
# 2000/07/20	v0.21, Joor Loohuis
#		changed code for regexes to remove qouble quotes when
#		reading, and putting them back when formatting
# 2000/07/14	v0.20, Joor Loohuis
#		changed the @EXPORT_OK to EXPORT, since it is unlikely
#		that the symbols will conflict with anything else
#		small corrections to the pod
# 2000/07/08	v0.13, Joor Loohuis
#		changed continuation character replacement to use 0x1E
#		rather than non breaking space
# 2000/05/16	v0.12, Joor Loohuis
#		adjusted Makefile.PL to generate man page
# 2000/03/31	v0.11, Joor Loohuis
#		fixed delimiter field reading
# 2000/03/27	v0.10, Joor Loohuis (joor@nl.linux.org)
#		original version
#
# ACKNOWLEDGEMENTS
#
# A large amount of the tricks employed here were copied or derived
# from the recipes in the Ram book, so thanks Tom and Nat


package NEdit::Config;

$VERSION = '0.23';

use strict;
use 5.005;

# report errors in module routines from code where the call came from
use Carp;

# preserve insertion order in hashes, important for data
# such as language patterns and user defined menus, and also 
# makes it simple to write configuration data in the correct order
use Tie::IxHash 1.21;

use vars qw(
    $VERSION @ISA @EXPORT
    $CONT $SMART_SEP $DEFAULT
    %DEFAULT_CONFIG @CONFIG_ITEMS
);

require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
    read_config
    format_config
    %DEFAULT_CONFIG
);


# use a less ambiguous replacement for line continuation chars
$CONT = "\x1E"; # record separator
# smartindent macro separator
$SMART_SEP = '--End-of-Macro--';
# some settings can have built-in content, signaled as such in the configuration
$DEFAULT = 'Default';


=pod

=head1 NAME

NEdit::Config - Configuration utilities for NEdit

=head1 SYNOPSIS

    # import required symbols
    use NEdit::Config;

    tie my %config, "Tie::IxHash";
    %config = read_config("$ENV{HOME}/.nedit");

or more tersely

    tie my %config, "Tie::IxHash", read_config("$ENV{HOME}/.nedit");


=head1 DESCRIPTION

NEdit::Config provides extensions to Perl for use in the manipulation of NEdit
configuration and import files. The core functionality is the reading of these
files and storing all information in a hash, and writing (partial)
configuration hashes to file. 

No symbols are exported into the namespace of the calling code, but in stead
subroutines and data must be imported explicitly. The symbols that can be
imported are listed in this section.

=head2 Configuration hash

The hash containing a NEdit configuration can contain any or all resources
available in the configuration files. The configuration hash is indexed with
the resource names, from which the 'nedit.' prefix is removed. Since a hash can
only contain scalar values, resources with non-scalar contents are stored by
reference. Due to the complex nature of the NEdit configuration, individual
resources may in turn contain references to hashes and arrays. The map below
describes the hierarchical structure for the configuration hash. Keys are
quoted if literal, and depend on the configuration otherwize.

  key    : 'fileVersion'
  type   : string
  content: NEdit version

  key    : 'shellCommands'
  type   : hash reference
  content: list of shell menu commands

  	key    : command name
	type   : hash reference
	content: shell menu command parameters

		key    : 'accellerator'
		type   : string
		content: shell command accellerator

		key    : 'mnemonic'
		type   : string
		content: shell menu mnemonic

		key    : 'io'
		type   : string
		content: command IO parameters (not processed)

		key    : 'command'
		type   : string
		content: string containing actual shell command

  key    : 'macroCommands'
  type   : hash reference
  content: macros

  	key    : macro menu entry
	type   : hash reference
	content: macro parameters

		key    : 'accellerator'
		type   : string
		content: macro accellerator

		key    : 'mnemonic'
		type   : string
		content: macro menu mnemonic

		key    : 'select'
		type   : string
		content: flag indicating required selection

		key    ; 'code'
		type   : array reference
		content: macro code

  			index  : line number - 1
			type   : string
			content: macro lines

  key    : 'bgMenuCommands'
  type   : hash reference
  content: background menu commands

  	key    : background menu entry
	type   : hash reference
	content: macro parameters

		key    : 'accellerator'
		type   : string
		content: macro accellerator

		key    : 'mnemonic'
		type   : string
		content: macro menu mnemonic

		key    : 'select'
		type   : string
		content: flag indicating required selection

		key    ; 'code'
		type   : array reference
		content: macro code

  			index  : line number - 1
			type   : string
			content: macro lines

  key    : 'highlightPatterns'
  type   : hash reference
  content: highlight pattern sets

  	key    : language name
	type   : hash reference or string (see below)
	content: highlight pattern definition

		key    : 'context lines'
		type   : string
		content: context requirement lines

		key    : 'context chars'
		type   : string
		content: context requirement chars

		key    : 'patterns'
		type   : hash reference
		content: individual highlight patterns

			key    : pattern name
			type   : hash reference
			content: individual highlight pattern

				key    : 'starting re'
				type   : string
				content: starting regular expression

				key    : 'ending re'
				type   : string
				content: ending regular expression

				key    : 'error re'
				type   : string
				content: error regular expression

				key    : 'style'
				type   : string
				content: pattern drawing style

				key    : 'parent'
				type   : string
				content: parent pattern name

				key    : 'type'
				type   : string
				content: pattern type

  key    : 'languageModes'
  type   : hash reference
  content: language modes

  	key    : language name
	type   : hash reference
	content: language mode specification

		key    : 'extensions'
		type   : string
		content: file name extensions

		key    : 'regex'
		type   : string
		content: recognition regular expression

		key    : 'indent'
		type   : string
		content: indent style

		key    : 'wrap'
		type   : string
		content: wrap style

		key    : 'hardware tab'
		type   : string
		content: alternative hardware tab size

		key    : 'emulated tab'
		type   : string
		content: alternative emulated tab size

		key    : 'delimiters'
		type   : string
		content: word delimiters

  key    : 'styles'
  type   : hash reference
  content: text drawing styles

  	key    : style name
	type   : hash reference
	content: style parameters

		key    : 'color'
		type   : string
		content: style color

		key    : 'font'
		type   : string
		content: style font

  key    : 'smartIndentInit'
  type   : hash reference
  content: smart indent macros

  	key    : language name
	type   : array reference
	content: smart indent macros

  		index  : line number - 1
		type   : string
		content: smart indent macro lines

  key    : 'smartIndentInitCommon'
  type   : array reference or string (see below)
  content: common smart indent initialization macro

  	index  : line number - 1
	type   : string
	content: smart indent common initialization macro lines

  key    : 'autoWrap'
  type   : string
  content: wrap style

  key    : 'wrapMargin'
  type   : string
  content: wrap margin

  key    : 'autoIndent'
  type   : string
  content: auto indent style

  key    : 'autoSave'
  type   : string
  content: incremental backup

  key    : 'saveOldVersion'
  type   : string
  content: make backup copy

  key    : 'showMatching'
  type   : string
  content: show matching (..)

  key    : 'highlightSyntax'
  type   : string
  content: syntax highlighting

  key    : 'searchDialogs'
  type   : string
  content: searching verbosity

  key    : 'retainSearchDialogs'
  type   : string
  content: keep search dialogs up

  key    : 'repositionDialogs'
  type   : string
  content: popups under pointer

  key    : 'sortOpenPrevMenu'
  type   : string
  content: sort open prev menu

  key    : 'statisticsLine'
  type   : string
  content: statistics line

  key    : 'iSearchLine'
  type   : string
  content: inremental search line

  key    : 'lineNumbers'
  type   : string
  content: show line numbers

  key    : 'warnFileMods'
  type   : string
  content: warn on files modified externally

  key    : 'warnExit'
  type   : string
  content: warn on exit

  key    : 'searchMethod'
  type   : string
  content: default search style

  key    : 'textRows'
  type   : string
  content: initial window height

  key    : 'textCols'
  type   : string
  content: initial window width

  key    : 'tabDistance'
  type   : string
  content: hardware tab spacing

  key    : 'emulateTabs'
  type   : string
  content: emulated tab spacing

  key    : 'insertTabs'
  type   : string
  content: use tab characters in padding and emulated tabs

  key    : 'textFont'
  type   : string
  content: primary text font

  key    : 'boldHighlightFont'
  type   : string
  content: font for bold highlighting

  key    : 'italicHighlightFont'
  type   : string
  content: font for italic highlighting

  key    : 'boldItalicHighlightFont'
  type   : string
  content: font for bold italic highlighting

  key    : 'smartTags'
  type   : string
  content: tag collision style

  key    : 'prefFileRead'
  type   : string
  content: ?

Highlight pattern hash entries contain either the string 'Default', or a
reference to a complex hash containing pattern data. Something similar is the
case for the smart indent macro entries, which contain the same string or an
array reference. The difference is easily told by checking the result of the
ref function on the value: true for a hash or array reference, and false in
case it is a string.

=head2 Default configuration

The module has the data of the default configuration built in. The
configuration hash that contains it can be imported into a script as
C<%DEFAULT_CONFIG>. 

=cut

# process the 'shellCommands' string, and splits the data into a hash
# returns a reference to the shellcommands hash
sub process_shellCommands ($) {


    # split separate shell commands into an array
    $_[0] =~ s/^(\s|$CONT)*//;
    $_[0] =~ s/\\n$//;
    my @commands = split(/\\n$CONT\s*/, $_[0]);

    # split command properties into a hash
    my ($entry, $accel, $mnem, $io, $command);
    tie my %commands, "Tie::IxHash";
    while (@commands) {
	# get the command parameters and the actual command, which
	# are on separate lines
	($entry, $accel, $mnem, $io) = split(/:/, shift @commands);
	$command = shift @commands;
        tie my %command_data, "Tie::IxHash",
    	    ('accellerator' => $accel,
	     'mnemonic'     => $mnem,
	     'io'           => $io,
	     'command'      => $command);
	$commands{$entry} = \%command_data;
    }
    # io is not yet split, but contains a number of letters, which correspond to
    # settings in the Shell Commands dialog:
    # empty - command input is none, command output to same window
    # I - command input from selection
    # A - command input from window
    # E - command input from either
    # D - command output to dialog
    # W - command output to new window
    # X - output replaces input (implies command output to same window)
    # S - save input file before executing command
    # L - reload file after executing command
    
    # stuff the results back where they came from
    $_[0] = \%commands;
}

# processes a macro string, and splits the data into a hash
# returns a reference to the macro hash
# works for macroCommands and bgMenuCommands
sub process_macros ($) {

    # split all lines into an array
    $_[0] =~ s/^(\s|$CONT)*//;
    $_[0] =~ s/\\n$//;
    my @lines = split(/\\n$CONT\t?/, $_[0]);
    
    # go over lines, and get macro data and code
    my ($entry, $accel, $mnem, $select);
    tie my %macros, "Tie::IxHash";
    while (my $line = shift @lines) {
	my @code = ();
    	($entry, $accel, $mnem, $select) = split(/:/, $line);

	do {
	    $line = shift @lines;
	    push @code, $line;
	} until ($line =~ /^\}/); # got all macro code
	pop @code; # remove trailing brace

        tie my %macro_data, "Tie::IxHash",
    	    ('accellerator' => $accel,
	     'mnemonic'     => $mnem,
	     'select'       => $select,
	     'code'         => \@code);
        $macros{$entry} = \%macro_data;
    }

    # stuff the results back where they came from
    $_[0] = \%macros;
}

# processes the highlightPatterns string, and splits the data into a hash
# returns a reference to the highlight patterns hash
sub process_highlightPatterns ($) {

    # split all lines into an array
    $_[0] =~ s/^\s*//;
    $_[0] =~ s/\\n$//;
    my @lines = split(/\\n$CONT\t?/, $_[0]);
    
    # go over lines, and get language data and patterns
    my ($language);
    tie my %patterns, "Tie::IxHash";
    while (my $line = shift @lines) {
    	($language, $line) = split(/:/, $line, 2);
	if ($line !~ /$DEFAULT/) { # language doesn't use builtin patterns
	    my ($context_lines, $context_chars) = ($line =~ /^(\d+):(\d+){/);
	    # opening curly brace is discarded
	    tie my %pat_lines, "Tie::IxHash";
	    tie my %pats, "Tie::IxHash",
	    	('context lines' => $context_lines,
		 'context chars' => $context_chars,
	    	 'patterns'      => \%pat_lines);

	    while (($line = shift @lines) !~ /^}/) { # process individual pattern lines
	    	$line =~ s/^\s*//;
		my ($name, $start_re, $end_re, $error_re, $style, $parent, $type) =
		    ($line =~ /([^:]+):"(.+)":(?:|"(.*)"):(?:|"(.*)"):([^:]+):([^:]*):([CD]*)/);
		# matching double quoted options it somewhat odd, because the result
		# must not be undef, but an empty string if the option is empty
		# (".*")? gives undef for an empty option
		tie my %pattern_data, "Tie::IxHash",
		    ('starting re' => $start_re,
		     'ending re'   => $end_re,
		     'error re'    => $error_re,
		     'style'       => $style,
		     'parent'      => $parent,
		     'type'        => $type);
		# 'type' is not split up any further, but can contain
		# C - coloring for subexpressions of parent pattern
		# D - deferred parsing
		$pat_lines{$name} = \%pattern_data;
	    }
	    # closing curly brace is discarded
	    
	    $patterns{$language} = \%pats;
	}
	else {$patterns{$language} = $DEFAULT}
    }

    # stuff the results back where they came from
    $_[0] = \%patterns;
}

# processes the languageModes string, and splits the data into a hash
# returns a reference to the language modes hash
sub process_languageModes ($) {

    # split all lines into an array
    $_[0] =~ s/^\s*//;
    $_[0] =~ s/\\n$//;
    my @lines = split(/\\n$CONT\t?/, $_[0]);
    
    # go over lines, and get language data
    tie my %modes, "Tie::IxHash";
    while (my $line = shift @lines) {
	my ($language, $extensions, $regex, $indent, $wrap, $hard_tab, $em_tab, $delims) =
	    ($line =~ /([^:]+):([^:]*):(?:|"(.*)"):([^:]*):([^:]*):(\d*):(\d*):(?:|"(.*)")$/);
	# matching double quoted options it somewhat odd, because the result
	# must not be undef, but an empty string if the option is empty
	# (".*")? gives undef for an empty option
	tie my %language_data, "Tie::IxHash",
	    ('extensions'   => $extensions,
	     'regex'        => $regex,
	     'indent'       => $indent,
	     'wrap'         => $wrap,
	     'hardware tab' => $hard_tab,
	     'emulated tab' => $em_tab,
	     'delimiters'   => $delims);
        $modes{$language} = \%language_data;
    }

    # stuff the results back where they came from
    $_[0] = \%modes;
}

# processes the 'styles' string, and splits the data into a hash
# returns a reference to the styles hash
sub process_styles ($) {

    # split separate styles into an array
    $_[0] =~ s/^\s*//;
    my @styles = split(/\\n$CONT\s*/, $_[0]);

    # split style properties into a hash
    my ($name, $color, $font);
    tie my %styles, "Tie::IxHash";
    foreach (@styles) {
    	($name, $color, $font) = split /:/;
	tie my %style_data, "Tie::IxHash";
	%style_data = ('color' => $color, 
	    	       'font'  => $font);
	$styles{$name} = \%style_data;
    }

    # stuff the results back where they came from
    $_[0] = \%styles;
}

# processes the 'smartIndentInit' string, and splits the data into a hash
# returns a reference to the smartindent macro hash
sub process_smartIndentInit ($) {

    # split all lines into an array
    $_[0] =~ s/^\s*//;
    $_[0] =~ s/\\n$//;
    my @lines = split(/\\n$CONT\t?/, $_[0]);

    # go over lines, and get language name and macro code
    my $language;
    tie my %macros, "Tie::IxHash";
    while (my $line = shift @lines) {
	my @code = ();
	my $count = 0;
    	($language, $line) = ($line =~ /\s*(.+):\s*(.*)$/);
	if ($line !~ /$DEFAULT/) { # language doesn't use builtin macros
            push @code, $line;
	    do {
	    	$line = shift @lines;
	    	push @code, $line;
		$count++ if ($line =~ /$SMART_SEP/);
	    } until ($count == 3); # got all macros
            $macros{$language} = \@code;
	}
	else {$macros{$language} = $DEFAULT};
    }
    # @code contains the code for the smart indent macros, separated by
    # lines containing $SMART_SEP, or the single line 'Default'

    # stuff the results back where they came from
    $_[0] = \%macros;
}

# processes the 'smartIndentInitCommon' string, and splits the lines into an array
# returns a reference to the macro code array
sub process_smartIndentInitCommon ($) {

    # split separate code lines into an array
    $_[0] =~ s/^\s*//;
    $_[0] =~ s/\\n$//;
    
    if ($_[0] =~ /^$DEFAULT/) {return}; # default macro

    my @code = split(/\\n$CONT\t?/, $_[0]);

    # stuff the results back where they came from
    $_[0] = \@code;
}

# process the data if it has multiple values, and
# simply return it otherwize
sub process ($$) {
    ($_, my $data) = @_;
    # below is a switch statement, in case you couldn't tell
       /shellCommands/    	&& {return process_shellCommands($data)}
    || /macroCommands/	    	&& {return process_macros($data)}
    || /bgMenuCommands/     	&& {return process_macros($data)}
    || /highlightPatterns/  	&& {return process_highlightPatterns($data)}
    || /languageModes/	    	&& {return process_languageModes($data)}
    || /styles/     	    	&& {return process_styles($data)}
    || /smartIndentInit\b/    	&& {return process_smartIndentInit($data)}
    || /smartIndentInitCommon/	&& {return process_smartIndentInitCommon($data)}
    ||   	    	    	   {return $data};
}

=head2 Subroutines

=over 4

=item read_config

Reads a NEdit configuration file (typically $HOME/.nedit), and processes it.
Takes a file name as the parameter, and returns a hash containing the
configuration.

=cut


sub read_config ($) {

    # check for a file name, and open it
    my $handle;
    if ($_[0] ne "") { 
    	open(CONFIG, $_[0]) or croak __PACKAGE__, ": can't open configuration file: $!\n";
	$handle = *CONFIG;
    }
    else { $handle = *DATA } # for internal use only

    # read the file
    my ($line, $key, $value);
    tie my %config, "Tie::IxHash";

LINE: while (defined ($line = <$handle>)) {

	# tack continued lines onto eachother
    	chomp $line;
	if ($line =~ s/\\$/$CONT/) { # replace line continuation char
	    $line .= <$handle>;
	    redo LINE unless eof $handle;
	}

	# trim various garbage
	$line =~ s/^\s*!.*//;  # strip comments,
	$line =~ s/^\s+//; # leading, and
	$line =~ s/\s+$//; # trailing whitespace
	next LINE unless length $line;

	# split line into hash
	($key, $value) = split(/:/, $line, 2);
	$key =~ s/^nedit\.//; # clean up key
	$value =~ s/^\s*//; # remove whitespace
    
	# process various parts of the configuration that have multiple values
	# replace their entries in the configuration hash
	$config{$key} = process($key, $value);
    }
    close $handle;

    return %config;
}


# format shell commands hash for output
sub format_shellCommands ($) {
    tie my %commands, "Tie::IxHash", %{$_[0]};
    my $result = "";

    # process all entries
    while (my ($command, $data) = each %commands) {
    	$result .= 
	    "\\\n" .
	    "\t$command:" .
	    "$data->{'accellerator'}:" .
	    "$data->{'mnemonic'}:" .
	    "$data->{'io'}:\\n\\\n" .
	    "\t\t$data->{'command'}\\n";
    }
    
    return $result;
}


# format macros and background menu commands for output
sub format_macros ($) {
    tie my %macros, "Tie::IxHash", %{$_[0]};
    my $result = "";

    # process all entries
    while (my ($macro, $data) = each %macros) {
    	$result .=
	    "\\\n" .
	    "\t$macro:" .
	    "$data->{'accellerator'}:" .
	    "$data->{'mnemonic'}:" .
	    "$data->{'select'}: {\\n\\\n" .
	    "\t" .
	    join("\\n\\\n\t", @{$data->{'code'}}) .
	    "\\n\\\n\t}\\n";
    }
    
    return $result;
}


# format highlight patterns for output
sub format_highlightPatterns ($) {
    tie my %patterns, "Tie::IxHash", %{$_[0]};
    my $result = "";

    # process all entries
    while (my ($language, $data) = each %patterns) {
    	$result .=
	    "$language:";
	if (not ref $data) { # default language, no patterns
	    $result .= "$data\\n\\\n\t";
	    next;
	}
	$result .=
	    "$data->{'context lines'}:" .
	    "$data->{'context chars'}\{\\n\\\n";
	# list all patterns here
	while (my ($pattern, $specs) = each %{$data->{'patterns'}}) {
	    $result .= "\t\t$pattern:\"" . $specs->{'starting re'} . "\":";
	    if ($specs->{'ending re'}) {
		$result .= '"' . $specs->{'ending re'} . '"';
	    }
	    $result .= ":";
	    if ($specs->{'error re'}) {
		$result .= '"' . $specs->{'error re'} . '"';
	    }
	    $result .= ":" . 
		$specs->{'style'} . ":" .
		$specs->{'parent'} . ":" .
		$specs->{'type'};
	    $result .= "\\n\\\n";
	}
	$result .= "\t}\\n\\\n\t";
    }
    $result =~ s/\\n\\\n\t$//; # chop off excess line continuation
    
    return $result;
}


# format highlight patterns for output
sub format_languageModes ($) {
    tie my %modes, "Tie::IxHash", %{$_[0]};
    my $result = "";

    # process all entries
    while (my ($language, $data) = each %modes) {
	$result .= "\t$language:" . $data->{'extensions'} . ":";
	if ($data->{'regex'}) {
	    $result .= '"' . $data->{'regex'} . '"';
	}
	$result .= ":" . 
	    $data->{'indent'} . ":" .
	    $data->{'wrap'} . ":" .
	    $data->{'hardware tab'} . ":" .
	    $data->{'emulated tab'} . ":";
	if ($data->{'delimiters'}) {
	    $result .= '"' . $data->{'delimiters'} . '"';
	}
	$result .= "\\n\\\n";
    }
    $result =~ s/\\n\\\n$//; # chop off excess line continuation
    
    return $result;
}


# format styles for output
sub format_styles ($) {
    tie my %styles, "Tie::IxHash", %{$_[0]};
    my $result = "";

    # process all entries
    while (my ($style, $data) = each %styles) {
    	$result .=
	    "\t$style:" .
    	    join(":", values %{$data}) .
	    "\\n\\\n";
    }
    $result =~ s/\\n\\\n$//; # chop off excess line continuation
    
    return $result;
}


# format smart indent macros for output
sub format_smartIndentInit ($) {
    tie my %macros, "Tie::IxHash", %{$_[0]};
    my $result = "";

    # process all entries
    while (my ($language, $code) = each %macros) {
    	$result .= "\t$language:";
	if (not ref $code) {$result .= "$DEFAULT\\n\\\n"}
	else {$result .= "\t" . join("\\n\\\n\t", @{$code}) . "\\n\\\n"}
    }
    $result =~ s/\\n\\\n$//; # chop off excess line continuation

    return $result;
}


# format smart indent common initialization macros for output
sub format_smartIndentInitCommon ($) {
    if (not ref $_[0]) {return $DEFAULT};
    my @code = @{$_[0]};
    my $result = "\t" . join("\\n\\\n\t", @code) . "\\n";

    return $result;
}


# generate properly formatted strings for complex configuration data
sub format_data ($$) {
    ($_, my $data) = @_;
       /shellCommands/          && {return format_shellCommands($data)}
    || /macroCommands/	    	&& {return format_macros($data)}
    || /bgMenuCommands/         && {return format_macros($data)}
    || /highlightPatterns/      && {return format_highlightPatterns($data)}
    || /languageModes/	    	&& {return format_languageModes($data)}
    || /styles/     	    	&& {return format_styles($data)}
    || /smartIndentInit\b/      && {return format_smartIndentInit($data)}
    || /smartIndentInitCommon/  && {return format_smartIndentInitCommon($data)}
    ||   	    	    	   {return $data};
}


=pod

=item format_config

Writes the contents of a configuration hash into a string, using the format
NEdit requires. The hash can be imcomplete, items are written in the same order
as the default configuration. The output is not identical to what NEdit writes,
because NEdit is not entirely consistent in its use of whitespace at the start
of lines in the configuration.

=back

=cut

# format a configuration hash, passed by reference, into a string
# the configuration hash can be incomplete, for example when it contains data
# intended for an import file
# items are written in the order that is used by NEdit itself, irrespect of
# the order in which they are provided
sub format_config ($) {
    my $format = "";
    if ((ref $_[0]) ne 'HASH') {
    	carp __PACKAGE__, ": value passed to format_config not a hash reference";
	return $format;
    }
    my %config = %{$_[0]};
    
    # process items in order NEdit uses
    my $item;
    foreach $item (@CONFIG_ITEMS) {
    	next if not exists $config{$item};
	$format .= "nedit.$item: ";
	# perl doesn't like me for matching strings in a void context
	# (switch structure) so do that in a separate routine
	$format .= format_data($item, $config{$item});
	$format .= "\n";
    }
    return $format;
}


# read the built-in default configuration
eval {
    tie %DEFAULT_CONFIG, "Tie::IxHash", read_config("");
    @CONFIG_ITEMS = keys %DEFAULT_CONFIG;
};

=head1 NOTES

=over 4

=item *

Most items in the NEdit configuration are sensitive to the order in which
they are listed in the configuration. To preserve this order, NEdit::Config
uses the Tie::IxHash module, which preserves the insertion order when using the
keys, values and each functions on a hash. If such a tied hash is copied, the
copy must also be tied to avoid losing the order. This is done by using the
Tie::IxHash module to tie hashes:

    tie my %hash, "Tie::IxHash";
    %hash = read_config;

=back

=head1 RESTRICTIONS

Currently comments in configuration files read by read_config must be on
separate lines, since there is no way of telling whether a '!' is part of the
actual resource. I'm not sure whether NEdit applies the same restrictions.

=head1 AUTHOR

Joor Loohuis, joor@nl.linux.org

Copyright (c) 2000 Joor Loohuis. All rights reserved. This program is free
software; you can redistribute it and/or modify it under the same terms as Perl
itself.

=head1 SEE ALSO

perl(1).

=cut

1;
__DATA__

! Preferences file for NEdit
!
! This file is overwritten by the "Save Defaults..." command in NEdit 
! and serves only the interactively setable options presented in the NEdit
! "Preferences" menu.  To modify other options, such as background colors
! and key bindings, use the .Xdefaults file in your home directory (or
! the X resource specification method appropriate to your system).  The
! contents of this file can be moved into an X resource file, but since
! resources in this file override their corresponding X resources, either
! this file should be deleted or individual resource lines in the file
! should be deleted for the moved lines to take effect.

nedit.fileVersion: 5.1
nedit.shellCommands: \
	spell:Alt+B:s:EX:\n\
		cat>spellTmp; xterm -e ispell -x spellTmp; cat spellTmp; rm spellTmp\n\
	wc::w:ED:\n\
		set wc=`wc`; echo $wc[1] "lines," $wc[2] "words," $wc[3] "characters"\n\
	sort::o:EX:\n\
		sort\n\
	number lines::n:AW:\n\
		nl -ba\n\
	make:Alt+Z:m:W:\n\
		make\n\
	expand::p:EX:\n\
		expand\n\
	unexpand::u:EX:\n\
		unexpand\n
nedit.macroCommands: \
	Complete Word:Alt+D::: {\n\
		# Tuning parameters\n\
		ScanDistance = 200\n\
		\n\
		# Search back to a word boundary to find the word to complete\n\
		startScan = max(0, $cursor - ScanDistance)\n\
		endScan = min($text_length, $cursor + ScanDistance)\n\
		scanString = get_range(startScan, endScan)\n\
		keyEnd = $cursor-startScan\n\
		keyStart = search_string(scanString, "<", keyEnd, "backward", "regex")\n\
		if (keyStart == -1)\n\
		    return\n\
		keyString = "<" substring(scanString, keyStart, keyEnd)\n\
		\n\
		# search both forward and backward from the cursor position.  Note that\n\
		# using a regex search can lead to incorrect results if any of the special\n\
		# regex characters is encountered, which is not considered a delimiter\n\
		backwardSearchResult = search_string(scanString, keyString, keyStart-1, \\\n\
		    	"backward", "regex")\n\
		forwardSearchResult = search_string(scanString, keyString, keyEnd, "regex")\n\
		if (backwardSearchResult == -1 && forwardSearchResult == -1) {\n\
		    beep()\n\
		    return\n\
		}\n\
		\n\
		# if only one direction matched, use that, otherwise use the nearest\n\
		if (backwardSearchResult == -1)\n\
		    matchStart = forwardSearchResult\n\
		else if (forwardSearchResult == -1)\n\
		    matchStart = backwardSearchResult\n\
		else {\n\
		    if (keyStart - backwardSearchResult <= forwardSearchResult - keyEnd)\n\
		    	matchStart = backwardSearchResult\n\
		    else\n\
		    	matchStart = forwardSearchResult\n\
		}\n\
		\n\
		# find the complete word\n\
		matchEnd = search_string(scanString, ">", matchStart, "regex")\n\
		completedWord = substring(scanString, matchStart, matchEnd)\n\
		\n\
		# replace it in the window\n\
		replace_range(startScan + keyStart, $cursor, completedWord)\n\
	}\n\
	Fill Sel. w/Char:::R: {\n\
		if ($selection_start == -1) {\n\
		    beep()\n\
		    return\n\
		}\n\
		\n\
		# Ask the user what character to fill with\n\
		fillChar = string_dialog("Fill selection with what character?", "OK", "Cancel")\n\
		if ($string_dialog_button == 2 || $string_dialog_button == 0)\n\
		    return\n\
		\n\
		# Count the number of lines in the selection\n\
		nLines = 0\n\
		for (i=$selection_start; i<$selection_end; i++)\n\
		    if (get_character(i) == "\\n")\n\
		    	nLines++\n\
		\n\
		# Create the fill text\n\
		rectangular = $selection_left != -1\n\
		line = ""\n\
		fillText = ""\n\
		if (rectangular) {\n\
		    for (i=0; i<$selection_right-$selection_left; i++)\n\
			line = line fillChar\n\
		    for (i=0; i<nLines; i++)\n\
			fillText = fillText line "\\n"\n\
		    fillText = fillText line\n\
		} else {\n\
		    if (nLines == 0) {\n\
		    	for (i=$selection_start; i<$selection_end; i++)\n\
		    	    fillText = fillText fillChar\n\
		    } else {\n\
		    	startIndent = 0\n\
		    	for (i=$selection_start-1; i>=0 && get_character(i)!="\\n"; i--)\n\
		    	    startIndent++\n\
		    	for (i=0; i<$wrap_margin-startIndent; i++)\n\
		    	    fillText = fillText fillChar\n\
		    	fillText = fillText "\\n"\n\
			for (i=0; i<$wrap_margin; i++)\n\
			    line = line fillChar\n\
			for (i=0; i<nLines-1; i++)\n\
			    fillText = fillText line "\\n"\n\
			for (i=$selection_end-1; i>=$selection_start && get_character(i)!="\\n"; \\\n\
			    	i--)\n\
			    fillText = fillText fillChar\n\
		    }\n\
		}\n\
		\n\
		# Replace the selection with the fill text\n\
		replace_selection(fillText)\n\
	}\n\
	Quote Mail Reply:::: {\n\
		if ($selection_start == -1)\n\
		    replace_all("^.*$", "\\\\> &", "regex")\n\
		else\n\
		    replace_in_selection("^.*$", "\\\\> &", "regex")\n\
	}\n\
	Unquote Mail Reply:::: {\n\
		if ($selection_start == -1)\n\
		    replace_all("(^\\\\> )(.*)$", "\\\\2", "regex")\n\
		else\n\
		    replace_in_selection("(^\\\\> )(.*)$", "\\\\2", "regex")\n\
	}\n\
	C Comments>Comment Out Sel.@C@C++:::R: {\n\
		selStart = $selection_start\n\
		selEnd = $selection_end\n\
		replace_range(selStart, selEnd, "/* " get_selection() " */")\n\
		select(selStart, selEnd + 6)\n\
	}\n\
	C Comments>C Uncomment Sel.@C@C++:::R: {\n\
		sel = get_selection()\n\
		selStart = $selection_start\n\
		selEnd = $selection_end\n\
		commentStart = search_string(sel, "/*", 0)\n\
		if (substring(sel, commentStart+2, commentStart+3) == " ")\n\
		    keepStart = commentStart + 3\n\
		else\n\
		    keepStart = commentStart + 2\n\
		keepEnd = search_string(sel, "*/", length(sel), "backward")\n\
		commentEnd = keepEnd + 2\n\
		if (substring(sel, keepEnd - 1, keepEnd == " "))\n\
		    keepEnd = keepEnd - 1\n\
		replace_range(selStart + commentStart, selStart + commentEnd, \\\n\
			substring(sel, keepStart, keepEnd))\n\
		select(selStart, selEnd - (keepStart-commentStart) - \\\n\
			(commentEnd - keepEnd))\n\
	}\n\
	C Comments>+ C++ Comment@C++:::R: {\n\
		replace_in_selection("^.*$", "// &", "regex")\n\
	}\n\
	C Comments>- C++ Comment@C++:::R: {\n\
		replace_in_selection("(^[ \\\\t]*// ?)(.*)$", "\\\\2", "regex")\n\
	}\n\
	C Comments>+ C Bar Comment 1@C:::R: {\n\
		if ($selection_left != -1) {\n\
		    dialog("Selection must not be rectangular")\n\
		    return\n\
		}\n\
		start = $selection_start\n\
		end = $selection_end-1\n\
		origText = get_range($selection_start, $selection_end-1)\n\
		newText = "/*\\n" replace_in_string(get_range(start, end), \\\n\
			"^", " * ", "regex") "\\n */\\n"\n\
		replace_selection(newText)\n\
		select(start, start + length(newText))\n\
	}\n\
	C Comments>- C Bar Comment 1@C:::R: {\n\
		selStart = $selection_start\n\
		selEnd = $selection_end\n\
		newText = get_range(selStart+3, selEnd-4)\n\
		newText = replace_in_string(newText, "^ \\\\* ", "", "regex")\n\
		replace_range(selStart, selEnd, newText)\n\
		select(selStart, selStart + length(newText))\n\
	}\n\
	Make C Prototypes@C@C++:::: {\n\
		if ($selection_start == -1) {\n\
		    start = 0\n\
		    end = $text_length\n\
		} else {\n\
		    start = $selection_start\n\
		    end = $selection_end\n\
		}\n\
		string = get_range(start, end)\n\
		nDefs = 0\n\
		searchPos = 0\n\
		prototypes = ""\n\
		staticPrototypes = ""\n\
		for (;;) {\n\
		    headerStart = search_string(string, \\\n\
			    "^[a-zA-Z]([^;#\\"'{}=><!/]|\\n)*\\\\)[ \\t]*\\n?[ \\t]*\\\\{", \\\n\
			    searchPos, "regex")\n\
		    if (headerStart == -1)\n\
			break\n\
		    headerEnd = search_string(string, ")", $search_end,"backward") + 1\n\
		    prototype = substring(string, headerStart, headerEnd) ";\\n"\n\
		    if (substring(string, headerStart, headerStart+6) == "static")\n\
			staticPrototypes = staticPrototypes prototype\n\
		    else\n\
		    			prototypes = prototypes prototype\n\
		    searchPos = headerEnd\n\
		    nDefs++\n\
		}\n\
		if (nDefs == 0) {\n\
		    dialog("No function declarations found")\n\
		    return\n\
		}\n\
		new()\n\
		focus_window("last")\n\
		replace_range(0, 0, prototypes staticPrototypes)\n\
	}\n
nedit.bgMenuCommands: \
	Undo:::: {\n\
		undo()\n\
	}\n\
	Redo:::: {\n\
		redo()\n\
	}\n\
	Cut:::R: {\n\
		cut_clipboard()\n\
	}\n\
	Copy:::R: {\n\
		copy_clipboard()\n\
	}\n\
	Paste:::: {\n\
		paste_clipboard()\n\
	}\n
nedit.highlightPatterns: C:Default\n\
	C++:Default\n\
	Java:Default\n\
	JavaScript:Default\n\
	Ada:Default\n\
	Fortran:Default\n\
	Pascal:Default\n\
	Lex:Default\n\
	Yacc:Default\n\
	Perl:Default\n\
	Python:Default\n\
	Awk:Default\n\
	Tcl:Default\n\
	Sh Ksh Bash:Default\n\
	Csh:Default\n\
	Makefile:Default\n\
	SGML HTML:Default\n\
	LaTeX:Default\n\
	PostScript:Default\n\
	SQL:Default\n\
	Matlab:Default\n\
	VHDL:Default\n\
	Verilog:Default\n\
	X Resources:Default\n\
	NEdit Macro:Default
nedit.languageModes: 	C:.c .h::::::".,/\\`'!|@#%^&*()-=+{}[]"":;<>?~"\n\
	C++:.cc .hh .C .H .i .cxx .hxx::::::".,/\\`'!|@#%^&*()-=+{}[]"":;<>?~"\n\
	Java:.java::::::\n\
	JavaScript:.js::::::\n\
	Ada:.ada .ad .ads .adb .a::::::\n\
	Fortran:.f .f77 .for::::::\n\
	Pascal:.pas .p .int::::::\n\
	Lex:.lex::::::\n\
	Yacc:.y::::::".,/\\`'!|@#%^&*()-=+{}[]"":;<>?~"\n\
	Perl:.pl .pm .p5:"^[ \\t]*#[ \\t]*!.*perl":::::\n\
	Python:.py:"^#!.*python":Auto:None:::\n\
	Tcl:.tcl::::::\n\
	Awk:.awk::::::\n\
	Sh Ksh Bash:.sh .bash .ksh .profile:"^[ \\t]*#[ \\t]*![ \\t]*/bin/(sh|ksh|bash)":::::\n\
	Csh:.csh .cshrc .login .logout:"^[ \\t]*#[ \\t]*![ \\t]*/bin/csh":::::\n\
	Makefile:Makefile makefile .gmk:::None:8:8:\n\
	SGML HTML:.sgml .sgm .html .htm:"\\<[Hh][Tt][Mm][Ll]\\>":::::\n\
	LaTeX:.tex .sty .cls .dtx .ins::::::\n\
	PostScript:.ps .PS .eps .EPS .epsf .epsi:"^%!":::::"/%(){}[]<>"\n\
	SQL:.sql::::::\n\
	Matlab:.m .oct .sci::::::\n\
	VHDL:.vhd .vhdl .vdl::::::\n\
	Verilog:.v::::::\n\
	X Resources:.Xresources .Xdefaults .nedit:"^[!#].*([Aa]pp|[Xx]).*[Dd]efaults":::::\n\
	NEdit Macro:.nm .neditmacro::::::
nedit.styles: 	Plain:black:Plain\n\
	Comment:gray20:Italic\n\
	Keyword:black:Bold\n\
	Storage Type:brown:Bold\n\
	String:darkGreen:Plain\n\
	String1:SeaGreen:Plain\n\
	String2:darkGreen:Bold\n\
	Preprocessor:RoyalBlue4:Plain\n\
	Preprocessor1:blue:Plain\n\
	Character Const:darkGreen:Plain\n\
	Numeric Const:darkGreen:Plain\n\
	Identifier:brown:Plain\n\
	Identifier1:RoyalBlue4:Plain\n\
	Subroutine:brown:Plain\n\
	Subroutine1:chocolate:Plain\n\
	Ada Attributes:plum:Bold\n\
	Label:red:Italic\n\
	Flag:red:Bold\n\
	Text Comment:SteelBlue4:Italic\n\
	Text Key:VioletRed4:Bold\n\
	Text Key1:VioletRed4:Plain\n\
	Text Arg:RoyalBlue4:Bold\n\
	Text Arg1:SteelBlue4:Bold\n\
	Text Arg2:RoyalBlue4:Plain\n\
	Text Escape:gray30:Bold\n\
	LaTeX Math:darkGreen:Plain
nedit.smartIndentInit: 	C:Default\n\
	C++:Default\n\
	Python:Default\n\
	Matlab:Default
nedit.smartIndentInitCommon: Default
nedit.autoWrap: Newline
nedit.wrapMargin: 0
nedit.autoIndent: Auto
nedit.autoSave: True
nedit.saveOldVersion: False
nedit.showMatching: True
nedit.highlightSyntax: False
nedit.searchDialogs: False
nedit.retainSearchDialogs: False
nedit.repositionDialogs: True
nedit.sortOpenPrevMenu: True
nedit.statisticsLine: False
nedit.iSearchLine: False
nedit.lineNumbers: False
nedit.warnFileMods: True
nedit.warnExit: True
nedit.searchMethod: Literal
nedit.textRows: 24
nedit.textCols: 80
nedit.tabDistance: 8
nedit.emulateTabs: 0
nedit.insertTabs: True
nedit.textFont: -adobe-courier-medium-r-normal--12-*-*-*-*-*-*
nedit.boldHighlightFont: -adobe-courier-bold-r-normal--12-*-*-*-*-*-*
nedit.italicHighlightFont: -adobe-courier-medium-o-normal--12-*-*-*-*-*-*
nedit.boldItalicHighlightFont: -adobe-courier-bold-o-normal--12-*-*-*-*-*-*
nedit.smartTags: True
nedit.prefFileRead: False
